// The libMesh Finite Element Library.
// Copyright (C) 2002-2022 Benjamin S. Kirk, John W. Peterson, Roy H. Stogner

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



#ifndef LIBMESH_CELL_PRISM20_H
#define LIBMESH_CELL_PRISM20_H

// Local includes
#include "libmesh/cell_prism.h"

namespace libMesh
{

/**
 * The \p Prism20 is an element in 3D composed of 20 nodes.
 * It is numbered like this:
 *
 * \verbatim
 *   PRISM20:
 *               5
 *               o
 *              /:\
 *             / : \
 *            /  :  \
 *           /   :   \
 *       14 o    :    o 13
 *         /     :     \
 *        /      :      \
 *       /       o 11    \
 *    3 /        :        \4
 *     o---------o---------o
 *     |         :12       |
 *     |         :         |
 *     |    o    :    o    |            zeta
 *     |   17    o   16    |             ^   eta (into page)
 *     |        .2.        |             | /
 *     |       .   .       |             |/
 *   9 o      .  o  .      o 10          o---> xi
 *     |     .  15   .     |
 *     |  8 o         o 7  |
 *     |   .           .   |
 *     |  .             .  |
 *     | .               . |
 *     |.                 .|
 *     o---------o---------o
 *     0         6         1
 * \endverbatim
 *
 * And it also includes two face nodes:
 * Node 18, on side 0, equidistant from 0/1/2 or 6/7/8
 * Node 19, on side 4, equidistant from 3/4/5 or 12/13/14
 *
 * (xi, eta, zeta): { 0  <= xi   <= 1
 *                  { 0  <= eta  <= 1
 *                  { -1 <= zeta <= 1
 *                  { xi + eta   <= 1
 * are the reference element coordinates associated with the given
 * numbering.
 *
 * \author Roy H. Stogner
 * \date 2022
 * \brief A 3D prismatic element with 20 nodes.
 */
class Prism20 final : public Prism
{
public:

  /**
   * Constructor.  By default this element has no parent.
   */
  explicit
  Prism20 (Elem * p=nullptr) :
    Prism(Prism20::n_nodes(), p, _nodelinks_data)
  {}

  Prism20 (Prism20 &&) = delete;
  Prism20 (const Prism20 &) = delete;
  Prism20 & operator= (const Prism20 &) = delete;
  Prism20 & operator= (Prism20 &&) = delete;
  virtual ~Prism20() = default;

  /**
   * \returns \p PRISM20.
   */
  virtual ElemType type () const override { return PRISM20; }

  /**
   * \returns 20.
   */
  virtual unsigned int n_nodes() const override { return num_nodes; }

  /**
   * \returns 8.
   */
  virtual unsigned int n_sub_elem() const override { return 8; }

  /**
   * \returns \p true if the specified (local) node number is a vertex.
   */
  virtual bool is_vertex(const unsigned int i) const override;

  /**
   * \returns \p true if the specified (local) node number is an edge.
   */
  virtual bool is_edge(const unsigned int i) const override;

  /**
   * \returns \p true if the specified (local) node number is a face.
   */
  virtual bool is_face(const unsigned int i) const override;

  /**
   * \returns \p true if the specified (local) node number is on the
   * specified side.
   */
  virtual bool is_node_on_side(const unsigned int n,
                               const unsigned int s) const override;

  virtual std::vector<unsigned int> nodes_on_side(const unsigned int s) const override;

  virtual std::vector<unsigned int> nodes_on_edge(const unsigned int e) const override;

  /**
   * \returns \p true if the specified (local) node number is on the
   * specified edge.
   */
  virtual bool is_node_on_edge(const unsigned int n,
                               const unsigned int e) const override;

  /**
   * \returns \p true if the element map is definitely affine within
   * numerical tolerances.
   */
  virtual bool has_affine_map () const override;

  /**
   * \returns SECOND.
   */
  virtual Order default_order() const override;

  /**
   * Don't hide Elem::key() defined in the base class.
   */
  using Elem::key;

  /**
   * \returns An id associated with the \p s side of this element.
   * The id is not necessarily unique, but should be close.  This is
   * particularly useful in the \p MeshBase::find_neighbors() routine.
   *
   * We reimplement this method here for the \p Prism20 since we can
   * use the center node of each quad face to provide a perfect (unique)
   * key.
   */
  virtual dof_id_type key (const unsigned int s) const override;

  /**
   * \returns \p Prism20::side_nodes_map[side][side_node] after doing some range checking.
   */
  virtual unsigned int local_side_node(unsigned int side,
                                       unsigned int side_node) const override;

  /**
   * \returns \p Prism20::edge_nodes_map[edge][edge_node] after doing some range checking.
   */
  virtual unsigned int local_edge_node(unsigned int edge,
                                       unsigned int edge_node) const override;

  /**
   * Builds a \p QUAD9 or \p TRI6 built coincident with face i.
   * The \p std::unique_ptr<Elem> handles the memory aspect.
   */
  virtual std::unique_ptr<Elem> build_side_ptr (const unsigned int i,
                                                bool proxy=false) override;

  /**
   * Rebuilds a \p QUAD9 or \p TRI6 built coincident with face i.
   */
  virtual void build_side_ptr (std::unique_ptr<Elem> & elem,
                               const unsigned int i) override;

  /**
   * Builds a \p EDGE3 or \p INFEDGE2 built coincident with edge i.
   * The \p std::unique_ptr<Elem> handles the memory aspect.
   */
  virtual std::unique_ptr<Elem> build_edge_ptr (const unsigned int i) override;

  /**
   * Rebuilds a \p EDGE3 built coincident with edges 0 to 2, or \p
   * INFEDGE2 built coincident with edges 3 to 5.
   */
  virtual void build_edge_ptr (std::unique_ptr<Elem> & edge, const unsigned int i) override;

  virtual void connectivity(const unsigned int sc,
                            const IOPackage iop,
                            std::vector<dof_id_type> & conn) const override;

  /**
   * \returns 2 for all edge nodes, 4 for quad face nodes, 6 for tri
   * face nodes.
   */
  virtual unsigned int n_second_order_adjacent_vertices (const unsigned int) const override;

  /**
   * \returns The element-local number of the \f$ v^{th} \f$ vertex
   * that defines the \f$ n^{th} \f$ second-order node.
   *
   * \note \p n is counted as depicted above, \f$ 6 \le n < 20 \f$.
   */
  virtual unsigned short int second_order_adjacent_vertex (const unsigned int n,
                                                           const unsigned int v) const override;

  /**
   * \returns The child number \p c and element-local index \p v of the
   * \f$ n^{th} \f$ second-order node on the parent element.  See
   * elem.h for further details.
   */
  virtual std::pair<unsigned short int, unsigned short int>
  second_order_child_vertex (const unsigned int n) const override;

  /**
   * Geometric constants for Prism20.
   */
  static const int num_nodes = 20;
  static const int num_sides = 5;
  static const int num_edges = 9;
  static const int num_children = 8;
  static const int nodes_per_side = 9;
  static const int nodes_per_edge = 3;

  /**
   * This maps the \f$ j^{th} \f$ node of the \f$ i^{th} \f$ side to
   * element node numbers.
   */
  static const unsigned int side_nodes_map[num_sides][nodes_per_side];

  /**
   * This maps the \f$ j^{th} \f$ node of the \f$ i^{th} \f$ edge to
   * element node numbers.
   */
  static const unsigned int edge_nodes_map[num_edges][nodes_per_edge];

  virtual void permute(unsigned int perm_num) override final;

  virtual void flip(BoundaryInfo *) override final;

#ifdef LIBMESH_ENABLE_AMR
  virtual
  const std::vector<std::pair<unsigned char, unsigned char>> &
  parent_bracketing_nodes(unsigned int,
                          unsigned int) const override
  {
    // We can't get any good bracketing nodes for child tri-center
    // nodes on the parent prism midplane.  No refining Prism20 for
    // us.
    libmesh_not_implemented_msg("Prism20 cannot be refined; use Prism21 instead.");
  }
#endif

  unsigned int center_node_on_side(const unsigned short side) const override final;

  ElemType side_type (const unsigned int s) const override final;

protected:

  /**
   * Data for links to nodes.
   */
  Node * _nodelinks_data[num_nodes];



#ifdef LIBMESH_ENABLE_AMR

  /**
   * Matrix used to create the elements children.
   */
  virtual Real embedding_matrix (const unsigned int,
                                 const unsigned int,
                                 const unsigned int) const override
  {
    // We can't get any good bracketing nodes for child tri-center
    // nodes on the parent prism midplane, so we can't auto-calculate
    // embedding matrices and couldn't use them if we did.
    libmesh_not_implemented_msg("Prism20 cannot be refined; use Prism21 instead.");
    return 0;
  }

  LIBMESH_ENABLE_TOPOLOGY_CACHES;

#endif // LIBMESH_ENABLE_AMR

  /**
   * Matrix that tells which vertices define the location
   * of mid-side (or second-order) nodes.  This matrix
   * handles only the second-order nodes that are unique
   * to \p Prism20.  All other second-order nodes are identical
   * with \p Prism15, and are therefore handled through a
   * matrix contained in \p cell_prism.C
   */
  static const unsigned short int _remaining_second_order_adjacent_vertices[5][4];
};

} // namespace libMesh

#endif // LIBMESH_CELL_PRISM20_H
